import sys
from nuclide.rational import Rational
from decimal import Decimal
from math import pi, sin, cos, ceil
import re
import cairo

def gcd(a, b):
    while b:
        a, b = b, a % b
    return a

def lcm(a, b):
    return (a * b) / gcd(a, b)

def rational_lcm(a, b):
    return Rational(lcm(a._n, b._n), gcd(a._d, b._d))

def rational_lcm_many(nums):
    result = None
    for num in nums:
        if result is None:
            result = num
        else:
            result = rational_lcm(num, result)
    if result is None:
        return 0
    else:
        return result
    
def polar_to_cartesian(r, theta):
    return r * cos(theta), r * -sin(theta)
        
class NuclidePicture(object):
    def __init__(self):
        self._wheels = []
    
    def add_wheel(self, rotation, stick_length):
        """rotations is the number of times the wheel will rotate in a 'tick',
        and should be a rational number. stick_length is the length in points
        of the stick following the wheel, and should be a float.
        """
        
        self._wheels.append((rotation, stick_length))
    
    _number_re = re.compile(r'-?\d+(?:\.\d+)?')
    def parse_data(self, data):
        last_num = None
        for match in self._number_re.finditer(data):
            num = match.group()
            if last_num is None:
                num = Rational.from_exact(Decimal(num))
                if num == Rational(0):
                    last_num = ''
                else:
                    last_num = Rational(1) / num
            else:
                if isinstance(last_num, str): # Kludge
                    last_num = None
                self.add_wheel(last_num, float(num))
                last_num = None
                
    def num_ticks(self):
        """Returns the number of ticks necessary to complete the drawing."""
        
        rotations = []
        for rotation, stick_length in self._wheels:
            if rotation is not None:
                rotations.append(abs(rotation))
        return rational_lcm_many(rotations)
        
    def radius(self):
        result = 0
        for rotation, stick_length in self._wheels:
            result += stick_length
        return result
    
    def draw(self, ctx, line_width=2.0, resolution=1000):
        radius = self.radius()
        line_width /= 2.0
        ctx.move_to(line_width + radius, line_width)
        ticks = self.num_ticks()
        for i in xrange(resolution):
            tick = ticks * Rational(i, resolution)
            x = y = 0
            last_angle = 0
            for rotation, stick_length in self._wheels:
                if rotation is None:
                    angle = 0
                else:
                    angle = ((tick/rotation) * (pi * 2))
                angle += last_angle
                last_angle = angle
                angle += pi / 2.0
                this_x, this_y = polar_to_cartesian(stick_length, angle)
                x += this_x
                y += this_y
            ctx.line_to(radius + x + line_width, radius + y + line_width)
        ctx.close_path()
    
    def scale(self, n):
        self._wheels = [(rotation, stick_length*n) 
                            for rotation, stick_length in self._wheels]

def main():
    picture = NuclidePicture()
    sys.stdout.write('Enter data: ')
    data = sys.stdin.readline()
    picture.parse_data(data)
    sys.stdout.write('Enter SVG filename, or press enter for no SVG: ')
    svgname = sys.stdin.readline().strip()
    sys.stdout.write('Enter PNG filename, or press enter for no PNG: ')
    pngname = sys.stdin.readline().strip()
    if svgname:
        surface = cairo.SVGSurface(svgname, picture.radius() * 2 + 2,
                                   picture.radius() * 2 + 2)
    else:
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32,
                                     int(ceil(picture.radius() * 2 + 2)),
                                     int(ceil(picture.radius() * 2 + 2)))
    ctx = cairo.Context(surface)
    ctx.set_line_width(2)
    picture.draw(ctx, 2.0, 10000)
    ctx.stroke()
    if pngname:
        surface.write_to_png(pngname)
    ctx.show_page()
    surface.finish()
    
if __name__ == '__main__':
    main()
